mod grpc;
use dashmap::DashMap;
use grpc::dummy;
use std::pin::Pin;
use tokio_stream::{wrappers::ReceiverStream, Stream, StreamExt};
use tonic::async_trait;
use tonic::metadata::MetadataValue;
use tonic::{transport::Server, Request, Response, Status};

use dummy::{
    test_service_server::TestService, Empty, SomeRequest, SomeResponse, StreamRequest,
    StreamResponse,
};

#[derive(Debug, Default)]
pub struct MyTestService {
    challenges: DashMap<String, String>,
}

#[async_trait]
impl TestService for MyTestService {
    async fn get_something(
        &self,
        request: Request<SomeRequest>,
    ) -> Result<Response<SomeResponse>, Status> {
        println!("Got a request from the client: {:?}", request.get_ref().what_client_sends);
        println!("Get something receied: {:?}", request.metadata().get("request-signature"));

        let reply = SomeResponse {
            what_server_sends: format!(
                "Hello from server, you said: {}",
                request.get_ref().what_client_sends
            ),
        };

        Ok(Response::new(reply))
    }

    async fn client_messages(
        &self,
        request: Request<tonic::Streaming<StreamRequest>>,
    ) -> Result<Response<Empty>, Status> {
        println!("Client started streaming... {:?}", request.metadata().get("authorization"));
        let mut stream = request.into_inner();

        while let Some(result) = stream.next().await {
            match result {
                Ok(msg) => {
                    println!("Client streamed message: {}", msg.what_client_sends);
                }
                Err(e) => {
                    println!("Stream error: {}", e);
                    return Err(Status::internal(format!("Error reading stream: {}", e)));
                }
            }
        }

        println!("Client finished streaming.");

        Ok(Response::new(Empty {}))
    }

    type ServerMessagesStream = Pin<Box<dyn Stream<Item = Result<StreamResponse, Status>> + Send>>;

    async fn server_messages(
        &self,
        _request: Request<Empty>,
    ) -> Result<Response<Self::ServerMessagesStream>, Status> {
        let (tx, rx) = tokio::sync::mpsc::channel(4);

        tokio::spawn(async move {
            let messages =
                vec!["Hello from the server!", "Here's another message!", "And one more message!"];

            for msg in messages {
                let response = StreamResponse { what_server_sends: msg.to_string() };

                if let Err(e) = tx.send(Ok(response)).await {
                    eprintln!("Failed to send response via stream: {e}");
                    return;
                }
            }
        });

        let output_stream = ReceiverStream::new(rx);

        Ok(Response::new(Box::pin(output_stream) as Self::ServerMessagesStream))
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let addr = "[::1]:50051".parse()?;
    let service = MyTestService::default();

    println!("Starting gRPC server on {}", addr);
    let service =
        dummy::test_service_server::TestServiceServer::with_interceptor(service, check_auth);

    Server::builder().add_service(service).serve(addr).await?;

    Ok(())
}

fn check_auth(req: Request<()>) -> Result<Request<()>, Status> {
    let token: MetadataValue<_> = "Bearer some-secret-token".parse().unwrap();
    println!("Request: {:?}", req);

    match req.metadata().get("authorization") {
        Some(t) if token == t => Ok(req),
        _ => Err(Status::unauthenticated("No valid auth token")),
    }
}
